//===----------- CfgV_Fernando.cpp - Printing prototype -------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it prints the control flow graph in the dot format.
//
// Date: May 6th, 2006
//
//===----------------------------------------------------------------------===//
//
// This file prints all the instructions in the given machine function in the
// dot format, so that it can be visualized using Graphviz. To use it,
// add the definition of FernandoCfgID to llvm/CodeGen/Passes.h, and add
// this pass as a required pass to the PNE class (PHIElimination.cpp).
// Eg.: addRequired(FernandoCfgID). Then, compile the modified files, and
// link them into llc: makellvm llc in the $LLVM_HOME/lib/CodeGen directory.
//
//===----------------------------------------------------------------------===//


#include "llvm/Function.h"
#include "llvm/CodeGen/Passes.h"                 // FernandoPassID
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/BasicBlock.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include <iostream>                              // cerr
#include <sstream>                               // ostringstream


using namespace llvm;

namespace {

  #define INST_LEN 20
  #define OPRT_LEN 4

  class CfgV_Fernando : public MachineFunctionPass {

    /// All target-specific information can be accessed through this interface.
    const TargetMachine *TM;

    /// Information about the bank of registers implemented by the actual
    /// machine can be accessed through this interface.
    const MRegisterInfo *RegInfo;

    public:
      virtual void getAnalysisUsage(AnalysisUsage &AU) const {
        AU.addRequired<EdgeLiveness_Fernando>();
        MachineFunctionPass::getAnalysisUsage(AU);
      }

      virtual const char *getPassName() const {
        return "Print machine instructions.";
      }

    private:
      /// This variable contains liveness analysis information.
      EdgeLiveness_Fernando *edge_liveness;

      /// runOnMachineFunction - Prints information about each basic block.
      bool runOnMachineFunction(MachineFunction &Fn);

      /// AllocateBasicBlock - For the given basic block, prints the name
      /// of the basic block, and information about each of the
      /// instructions.
      void PrintBasicBlock(MachineBasicBlock &MBB);

      /// Prints the virtual registers used and defined by the instruction.
      void PrintOperands(MachineInstr *MI);

      /// Prints the name of the register.
      int PrintRegister(MachineOperand &MO);

      /// Prints the name, number, and neat header for the basic block.
      void PrintEdges(MachineBasicBlock &MBB);

      /// Fills str with n characters like 'c', where n = size(str) + packSize.
      /// It returns the number of characters printed.
      int PrintPackedStr(const char * str, int packSize, char c);
      int PrintPackedStr(std::string str, int packSize, char c);

      /// This method prints commands to format nodes and edges.
      void printHeaderInformation();

      /// Gives the name of the basic block
      std::string getBBName(MachineBasicBlock &MBB) {
        const BasicBlock *BB = MBB.getBasicBlock();
        std::ostringstream aux;
        aux << BB->getName();
        return aux.str();
      }
  };

  RegisterPass<CfgV_Fernando> W("FERNANDO_print_cfg", "Fernando print CFG.");

}


//===--------------------------------------------------------------------------
// The external declaration of FernandoCfgID is in llvm/CodeGen/Passes.h
// This contant is necessary to invoke this pass. For instance, if you
// want to invoke this pass before the phi elimination pass, on the
// PHIElimination.cpp file, add "AU.addRequiredID(FernandoCfgID);" in the
// "getAnalysisUsage" function.
//===--------------------------------------------------------------------------
const PassInfo *llvm::FernandoCfgID = W.getPassInfo();


//===--------------------------------------------------------------------------
// This function prints a basic block of the control flow graph. A basic block
// consists in a sequence of instructions, and a list of successors.
//===--------------------------------------------------------------------------
void CfgV_Fernando::PrintBasicBlock(MachineBasicBlock &MBB) {
  PrintEdges(MBB);
  std::string blockName = getBBName(MBB);
  std::cerr << "\"" << blockName << "\"" << " [ label=\"";

  // loop over each instruction
  MachineBasicBlock::iterator MII = MBB.begin();
  const TargetInstrInfo &TII = *TM->getInstrInfo();
  while (MII != MBB.end()) {
    MachineInstr *MI = MII++;
    const TargetInstrDescriptor &TID = TII.get(MI->getOpcode());
    PrintPackedStr(TID.Name, INST_LEN, ' ');
    PrintOperands(MI);
    std::cerr << "  \\l";
  }
  std::cerr << "\"]\n";
}


//===--------------------------------------------------------------------------
// This method prints the given string "str" and fills an amount of space
// equals to "packSize - str.size()" with the character "c". For instance,
// PrintPackedStr("Mov", 10, "-") results in "Mov-------".
//===--------------------------------------------------------------------------
int CfgV_Fernando::PrintPackedStr(std::string str, int packSize, char c) {
  int strSize = str.size();
  int incSize = strSize < packSize ? packSize - strSize : 0 ;
  std::string strAux(incSize, c);
  std::cerr << str << strAux;
  return (strSize > packSize) ? strSize : packSize ;
}


//===--------------------------------------------------------------------------
// This method prints instructions in a tabular format. Each instruction is
// printed in the format: INST_NAME    r1 = r2, ... rn, where r1 is the
// defined register, and r_i, i > 1 are the used registers. To change the
// amount of space between instructions and registers, set the macro
// INST_LEN. To change the amount of space between registers, set the macro
// OPRT_LEN. 
//===--------------------------------------------------------------------------
int CfgV_Fernando::PrintPackedStr(const char * str, int packSize, char c) {
  int strSize = strlen(str);
  int incSize = strSize < packSize ? packSize - strSize : 0 ;
  std::string strAux(incSize, c);
  std::cerr << str << strAux;
  return (strSize > packSize) ? strSize : packSize ;
}


//===--------------------------------------------------------------------------
// This method prints all the edges from the block "MBB" to all its
// successors. Edges are augmented with liveness analysis information.
// The label of each edge consists on the set of registers that are
// alive at that edge.
//===--------------------------------------------------------------------------
void CfgV_Fernando::PrintEdges(MachineBasicBlock &MBB) {
  const BasicBlock *BB = MBB.getBasicBlock();
  std::string bbName = "\"" + getBBName(MBB) + "\"";

  // Prints all the successors:
  for(MachineBasicBlock::succ_iterator succ = MBB.succ_begin();
      succ != MBB.succ_end();
      succ++) {
    MachineBasicBlock* mbb = *succ;
    std::string label;
    std::ostringstream label_aux;
    label_aux << "[label = \"" << edge_liveness->to_string(MBB, *mbb) << "\"]";
    label = label_aux.str();
    const BasicBlock *bb = mbb->getBasicBlock();
    std::cerr << bbName << " -> " << "\"" << bb->getName() << "\"" << " "
              << label << "\n";
  }
}


//===--------------------------------------------------------------------------
// This method prints a register. The register is packed with white spaces,
// i.e (' '), and the amount of space used is returned. Virtual registers
// are printed as a number prefixed by the letter 't'. Physical registers
// have their full name printed.
//===--------------------------------------------------------------------------
int CfgV_Fernando::PrintRegister(MachineOperand &MO) {
  int returnValue = 0;
  std::ostringstream aux;
  std::string result;
  if (MO.getReg()) {
    if(MRegisterInfo::isPhysicalRegister(MO.getReg())) {
      aux << RegInfo->getName(MO.getReg());
    }
    else {
      aux << "t" << (MO.getReg() - MRegisterInfo::FirstVirtualRegister);
    }
  }
  result = aux.str();
  returnValue = PrintPackedStr(result, OPRT_LEN, ' ');
  return returnValue;
}


//===--------------------------------------------------------------------------
// This method prints all the operands of an instruction via multiple calls
// of the method PrintRegister. Operands are separated by commas, and are
// printed in a tabular sequence filled with white spaces.
//===--------------------------------------------------------------------------
void CfgV_Fernando::PrintOperands(MachineInstr *MI) {
  bool isFirst = true;
  int column = INST_LEN;
  int colChange = 0;
  // first, print all the definitions
  for (unsigned i = 0; i != MI->getNumOperands(); ++i) {
    MachineOperand& MO = MI->getOperand(i);
    if (MO.isDef() && MO.isRegister() && MO.getReg()) {
      if(isFirst) {
        colChange += PrintRegister(MO);
        isFirst = false;
      } else {
        std::cerr << ", ";
        colChange += PrintRegister(MO);
      }
    }
  }
  if(colChange == 0) {
    PrintPackedStr("", OPRT_LEN, ' ');
  }
  // next, print all the uses
  std::cerr << " := ";
  isFirst = true;
  for (unsigned i = 0; i != MI->getNumOperands(); ++i) {
    MachineOperand& MO = MI->getOperand(i);
    if (MO.isUse() && MO.isRegister() && MO.getReg()) {
      if(isFirst) {
        PrintRegister(MO);
        isFirst = false;
      } else {
        std::cerr << ", ";
        PrintRegister(MO);
      }
    }
  }
}


//===--------------------------------------------------------------------------
// This method prints the header information of the dot graph. This information
// defines the font used for nodes and edges, plus the shape of nodes. In this
// case, nodes have rectangular shape, and the font is "Courier".
//===--------------------------------------------------------------------------
void CfgV_Fernando::printHeaderInformation() {
  // first prints he format of each node.
  std::cerr << "digraph foo {\n\n";
  std::cerr << "node [\n";
  std::cerr << "  fontname = \"Courier\"\n";
  std::cerr << "  fontsize = \"10\"\n";
  std::cerr << "  shape = \"box\"\n";
  std::cerr << "]\n\n";

  // second prints he format of each edge.
  std::cerr << "edge [\n";
  std::cerr << "  fontname = \"Courier\"\n";
  std::cerr << "  fontsize = \"8\"\n";
  std::cerr << "]\n\n";
}


//===--------------------------------------------------------------------------
// This method renders a visual representation of the data flow graph of
// function "Fn", in the dot format. This output can be graphically
// visualized with tools like Graphviz or Omnigraffle.
//===--------------------------------------------------------------------------
bool CfgV_Fernando::runOnMachineFunction(MachineFunction &Fn) {

  edge_liveness = &getAnalysis<EdgeLiveness_Fernando>();

  printHeaderInformation();

  // prints the label (title) of the graph:
  const Function *F = Fn.getFunction();
  std::cerr << "label=\"CFG for function " << F->getName() << "\"\n\n";

  // prints each basic block:
  TM = &Fn.getTarget();
  RegInfo = TM->getRegisterInfo();
  for (MachineFunction::iterator MBB = Fn.begin(), MBBe = Fn.end();
    MBB != MBBe; ++MBB)
    PrintBasicBlock(*MBB);

  std::cerr << "}\n";
  return false;
}
