//===----------- CfgV_Fernando.cpp - Printing prototype -------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it performs register allocation via choloring of chordal
// graphs.
//
// Date: June 7th, 2006
//
//===----------------------------------------------------------------------===//
//
// This file contains the implementation of a register allocation algorithm.
// Contrary to traditional RA approaches, this algorithm performs register
// assignment while the program is in SSA form, and uses parallel copies to
// eliminate phi functions, after a mapping between temporary and machine
// registers has been found. This algorithm is non-iterative, that is, it finds
// an optimal coloring of the interference graph in one scan of the control
// flow.
//
//===----------------------------------------------------------------------===//


#include <vector>
#include <sstream>                               // ostringstream (RegClique)
#include <iostream>                              // cerr

#include "llvm/CodeGen/Passes.h"                 // RegAllocChord_Fer
#include "llvm/Instructions.h"                   // MachineOperand, etc
#include "llvm/Analysis/LoopInfo.h"              // DefUseSites depends on it.
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/ADT/DepthFirstIterator.h"         // df_ext_iterator

#include "llvm/fernando/EdgeLiveness_Fernando.h"
#include "llvm/fernando/DefUseSites_Fer.h"
#include "llvm/fernando/RegClique_Fer.h"
#include "llvm/fernando/RegMapping_Fer.h"
#include "llvm/fernando/NeatPrint_Fer.h"

using namespace llvm;

namespace {

    class RegAllocChordal_Fer : public MachineFunctionPass {

    public:

    virtual void getAnalysisUsage(AnalysisUsage & AU) const {
        AU.addRequired<DefUseSites_Fer>();
        AU.addRequired<EdgeLiveness_Fernando>();
        MachineFunctionPass::getAnalysisUsage(AU);
    }

    /// The label of this pass is "Register allocation via coloring of
    /// chordal graphs."
    virtual const char * getPassName() const {
        return "Register allocation via coloring of chordal graphs.";
    }

    private:

    /// This interface represents the machine function that is beeing
    /// compiled. Different data structures can be accessed thrugh this
    /// interface, e.g:
    /// const TargetMachine & tm = this->machine_function.getTarget();
    MachineFunction * machine_function;


    /// This variable contains liveness analysis information.
    EdgeLiveness_Fernando * edge_liveness;


    /// This variable maps virtual registers to the instructions where it
    /// is defined and used.
    DefUseSites_Fer * def_use_sites;


    /// This data type keeps track of the mapping between virtual and
    /// physical registers.
    RegMapping_Fer * reg_mapping;


    /// This variable records the number of virtual registers in use in
    /// the target program.
    unsigned num_virtual_registers;


    /// This set contains the basic blocks that have already been visited
    /// by the register allocator.
    std::vector<bool> colored;


    /// After a virtual register is last used, the machine register assigned
    /// to it must be liberated, so it can be used to allocate other virtual
    /// registers. This method does exaclty this.
    void liberate_uses(RegClique_Fer & q, MachineInstr & mi);


    /// It is possible that the code being compiled contains dead definitions,
    /// that is, a definition of a variable that is never used after been
    /// defined. In this case, the machine register used to hold the value
    /// of such definition can be liberated to allocate other live ranges.
    void liberate_dead_defs(RegClique_Fer & q, MachineInstr & mi);


    /// This method handles the definition of pre-colored registers. When a
    /// pre-colored register p_reg is found, the following is done: if p_reg is
    /// empty, simply remove it from the list of available registers. If it
    /// contains a virtual register v_reg, than find a new location to v_reg,
    /// and handle p_reg as if it was empty.
    void allocate_physical_register(unsigned p_reg, MachineInstr & mi);


    /// This method is used to find a new location to the virtual v_reg. It
    /// is called when the register p_reg that v_reg currently uses needs to be
    /// used to store a different value, while v_reg is still alive. This may
    /// happen, for example, if p_reg is a precolored register, in the first
    /// instruction that defines p_reg. If there are other physical registers
    /// available, v_reg will have its value transfered to that register;
    /// otherwise, it will be spilled. The move instruction will be inserted
    /// before instruction mi.
    void move_or_store(MachineInstr & mi, unsigned v_reg, unsigned p_reg);


    /// This method gets one clique of register, and one instruction. The
    /// clique is supposed to represent all the registers alive immediately
    /// before the instruction is executed. The method assigns a valid
    /// physical register to the virtual register. If there is not physical
    /// register available, it will spill one of the virtual registers, to
    /// make room for the virtual. Before doing the allocation, the method
    /// verify if all the virtual registers in the given clique are
    /// allocated. It will be an error if any of them is not assigned to a
    /// physical register.
    void allocate_defs(MachineInstr & mi, std::vector<RegClique_Fer *> & ql);


    /// When a registers is spilled, it is necessary to insert load and store
    /// instructions to preserve the semantics of the control flow graph. This
    /// can be done in many different ways. The algorithm below is the "dummy"
    /// one: it inserts one store after the definition of the spilled value,
    /// and it inserts a load instruction before each use of that value.
    /// Because of this, the spilled virtual register is not alive in the edges
    /// between basic blocks, and its live range can safely be removed. This
    /// algorithm preserves the SSA properties of the original graph, because
    /// each new load defines a different temporary. The
    /// place_store_instruction algorithm is given below. Remember that each
    /// temporary has a unique definition point, due to the SSA property.
    void place_store_instruction(const MachineOperand & mo);


    /// Dummy placement of load instructions. Insert one load before each use
    /// of the spilled register.
    void place_load_instructions(const MachineOperand & mo);


    /// This method determines a candidate to be sent to memory among the
    /// registers stored in q. The heuristic function used will be
    /// given by the algorithm:
    /// Instruction def_i := definition_site(v);
    ///
    /// Unsigned w := 1 + 10 ** loop_info.nested_depth(def_i);
    /// 
    /// foreach ( Instruction i in use_sites(v) )
    /// |
    /// |    w := w + 1 + 10 ** loop_info.nested_depth(i);
    /// |
    /// end foreach
    ///
    /// v2w(v) := w;
    const MachineOperand * choose_register_to_spill(MachineOperand & mo);


    /// This is the dual of the initialize_data_structures method. It will
    /// free all the memory used during register allocation.
    void delete_data_structures();


    /// This method initializes the data structures used during register
    /// allocation. 
    void initialize_data_structures(MachineFunction &mf);


    /// runOnMachineFunction - Invokes the register allocation algorithm.
    bool runOnMachineFunction(MachineFunction & machine_function);


    /// This method finds a mapping between virtual registers and
    /// machine registers.
    void perform_register_assignment( MachineFunction & mf );


    /// This method replace phi functions by parallel copies. Parallel
    /// copies are implemented with a mix of move and xor instructions
    void eliminate_phi_functions( MachineFunction & mf );


    /// This method performs register spilling. The parameter indicates a
    /// class of registers. The spilling algorithm must evict a register from
    /// the same class as the parameter. The method implements the following
    /// algorithm:
    ///
    ///    unsigned v_sp = choose_register_to_spill(v_reg);
    ///    place_store_instruction(v_sp);
    ///    place_load_instructions(v_sp);
    ///    kill_live_range(v_sp);
    ///
    /// The quality of the machine code produced depends ultimately on how the
    /// spilled register is chosen, and on how the placement of load
    /// instructions is implemented.
    const MachineOperand * spill(MachineOperand & mo);


    /// If liveness analysis information is available, the set of cliques in a
    /// basic block can be found with a backwards linear scan. The first clique
    /// will be the out set of the basic block. Each time a new instruction is
    /// scanned, the register that it defines is removed from the clique, and
    /// any register that it uses, if not already in the clique, is added to
    /// it. Notice that phi instructions do not contribute uses for clique
    /// construction. Only registers defined in the phi function are taken into
    /// consideration.
    void RegAllocChordal_Fer::build_list_cliques
                 (std::vector<RegClique_Fer *> & cl, MachineBasicBlock & mbb);


    /// This method updates the set of registers with data from the given
    /// instruction. Any defined register is removed from the clique, and
    /// any used register is added to the clique. If the used register is
    /// already in the clique, nothing is done.
    RegClique_Fer * update_list_of_cliques
           (MachineBasicBlock & mbb, RegClique_Fer & q, MachineInstr & mi);


    /// This method finds a mapping between virtual and machine registers
    /// in the basic block. The colored set points which nodes have
    /// already been visited in the depth first traversal of the
    /// dominance tree.
    void perform_local_allocation(MachineBasicBlock & mbb);


    /// This method converts the mapping from virtual register to physical
    /// register found by the perform_local_allocation method to a real
    /// assignment of physical registers in each operand of each instruction
    /// of the given basic block. Actually, it would be possible to implement
    /// this method directly in the perform_local_allocation function, but the
    /// separate implementation makes the program more readable.
    void hard_code_register_mapping(MachineBasicBlock & mbb);


    /// This method returns true if the given basic block is contained
    /// in the given set, and false otherwise.
    bool has_already_been_visited(MachineBasicBlock & mbb);

    };

    RegisterPass<RegAllocChordal_Fer> Z("RegAllocChord_Fer",
    "Register Allocation via Coloring of Chordal Graphs.");

}


// the external declaration of RegAllocChord_Fer is in llvm/CodeGen/Passes.h
const PassInfo *llvm::RegAllocChord_Fer = Z.getPassInfo();


//===--------------------------------------------------------------------------
// This method returns true if the given basic block is contained in the given
// set, and false otherwise.
//===--------------------------------------------------------------------------
bool RegAllocChordal_Fer::has_already_been_visited(MachineBasicBlock & mbb) {
    return this->colored[mbb.getNumber()];
/*
    // visiting predecessors:

    for(MachineBasicBlock::pred_iterator pred_aux = mbb.pred_begin();
        pred_aux != mbb.pred_end(); pred_aux++) {
        MachineBasicBlock* pred_mbb = *pred_aux;
        if( has_already_been_visited(*pred_mbb) ) {
            std::cerr << " <- MBB_" << pred_mbb->getNumber() << "\n";
        }
    }
*/
}


//===--------------------------------------------------------------------------
// Traditional algorithms used to destroy phi functions normally replace them
// by copy instructions. Although the classical algorithms preserve the
// semantics of the SSA-form program, they may increase the number of registers
// found during the register assignment phase. To avoid this problem, we
// replace phi functions by parallel copies. Parallel copies are normally
// implemented with xor instructions (Remember the trick of swaping two
// variables without an auxiliary using xor?). The register swap is a costly
// operation, because it demands three instructions to be implemented;
// therefore, it should be avoided whenever possible. One way of avoiding
// swaps is by means of move instructions. Basically, swap operations are not
// necessary in every time that phi instructions are destroyed. This happens
// when the definition vector contains machine registers not used in the
// parameter vector. The algorithm shown below tries to use such optimization
// before inserting swap operations. It performs phi elimination in time
// linear in the size of the definition vector.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::eliminate_phi_functions( MachineFunction & mf ) {
}


//===--------------------------------------------------------------------------
// This method initializes the data structures used during register allocation.
// Among the variables initilized, are num_virtual_registers, edge_liveness,
// and def_use_sites.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::initialize_data_structures(MachineFunction & mf) {
    // finds the number of virtual registers in the target function
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned first_v_reg = MRegisterInfo::FirstVirtualRegister;
    this->num_virtual_registers = last_v_reg - first_v_reg;

    // gets the liveness analysis information
    this->edge_liveness = & getAnalysis<EdgeLiveness_Fernando>();

    this->def_use_sites = & getAnalysis<DefUseSites_Fer>();
    this->def_use_sites->dump();

    // gets information about target architecture
    this->machine_function =  & mf;

    // initializes the vector that marks visited nodes.
    MachineBasicBlock * last_block = mf.getLastBlock();
    unsigned num_blocks = last_block->getNumber() + 1;
    this->colored.assign(num_blocks, false);

    // initializes the mapping between virtual and physical registers.
    this->reg_mapping = new RegMapping_Fer(mf);
}


//===--------------------------------------------------------------------------
// This is the dual of the initialize_data_structures method. It will free
// all the memory used during register allocation.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::delete_data_structures() {
    // initializes the mapping between virtual and physical registers.
    delete this->reg_mapping;
}


//===--------------------------------------------------------------------------
// When a registers is spilled, it is necessary to insert load and store
// instructions to preserve the semantics of the control flow graph. This can
// be done in many different ways. The algorithm below is the "dummy" one: it
// inserts one store after the definition of the spilled value, and it inserts
// a load instruction before each use of that value. Because of this, the
// spilled virtual register is not alive in the edges between basic blocks, and
// its live range can safely be removed. This algorithm preserves the SSA
// properties of the original graph, because each new load defines a different
// temporary. The place_store_instruction algorithm is given below. Remember
// that each temporary has a unique definition point, due to the SSA property.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::place_store_instruction(const MachineOperand & mo) {
    unsigned v_reg = mo.getReg();
    if( ! this->def_use_sites->has_already_been_sent_to_memory(mo) ) {
        const MachineInstr * mi = this->def_use_sites->get_def_site(mo);
        const MachineBasicBlock * mbb = mi->getParent();
        // add store instruction
        // add load instruction before the end of the basic block.
        std::cerr << "Adding store after " <<
                  NeatPrint_Fer::print_instr(* mi, * this->machine_function) <<
                                " in basic block " << mbb->getNumber() << "\n";
        if(reg_mapping->get_physical_location(v_reg)!=RegMapping_Fer::NO_REG) {
            std::cerr << "The spilled register has already been colored\n";
        }
    }
}


//===--------------------------------------------------------------------------
// Dummy placement of load instructions. Insert one load before each use of
// the spilled register.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::place_load_instructions(const MachineOperand & mo) {
    unsigned v_reg = mo.getReg();
    if( ! this->def_use_sites->has_already_been_sent_to_memory(mo) ) {
        const std::vector< const MachineInstr * > use_sites =
                                        this->def_use_sites->get_use_sites(mo);
        for(unsigned u = 0; u < use_sites.size(); u++) {
            const MachineInstr * mi = use_sites[u];
            const MachineBasicBlock * mbb = mi->getParent();
            std::cerr << "Adding load before " <<
                 NeatPrint_Fer::print_instr(* mi, * this->machine_function) <<
                                " in basic block " << mbb->getNumber() << "\n";
            if(reg_mapping->get_physical_location(v_reg)
                                                   != RegMapping_Fer::NO_REG) {
                std::cerr << "The spilled register has already been colored\n";
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method determines a candidate to be sent to memory.
//===--------------------------------------------------------------------------
const MachineOperand * RegAllocChordal_Fer::choose_register_to_spill
                                                        (MachineOperand & mo) {
    unsigned v_reg = mo.getReg();
    std::vector<unsigned> usefull_regs;

    // Take the register class.
    const TargetRegisterClass *trc =
                    this->machine_function->getSSARegMap()->getRegClass(v_reg);
    TargetRegisterClass::iterator trci =
                         trc->allocation_order_begin(* this->machine_function);
    TargetRegisterClass::iterator trce =
                           trc->allocation_order_end(* this->machine_function);

    for (; trci != trce; ++trci) {
        unsigned v_reg = this->reg_mapping->get_assigned_data(*trci);
        usefull_regs.push_back(v_reg);
    }

    unsigned best_cdd = RegMapping_Fer::NO_REG;
    unsigned best_w = 65535;
    for(unsigned u = 0; u < usefull_regs.size(); u++) {
        unsigned v_reg = usefull_regs[u];
        unsigned w = this->def_use_sites->get_weight(v_reg);
std::cerr << "Reg = t" << (v_reg - MRegisterInfo::FirstVirtualRegister) << ", w = " << w << "\n";
        if(w < best_w) {
            best_cdd = v_reg;
            best_w = w;
        }
    }

std::cerr << "Best candidate = t" << (best_cdd - MRegisterInfo::FirstVirtualRegister) << ", w = " << best_w << "\n";
    return this->def_use_sites->get_machine_operand(best_cdd);
}


//===--------------------------------------------------------------------------
// This method performs register spilling. The parameter indicates a class of
// registers. The spilling algorithm must evict a register from the same class
// as the parameter. The method implements the following algorithm:
//
//    unsigned v_sp = choose_register_to_spill(v_reg);
//    place_store_instruction(v_sp);
//    place_load_instructions(v_sp);
//    kill_live_range(v_sp);
//
// The quality of the machine code produced depends ultimately on how the
// spilled register is chosen, and on how the placement of load instructions is
// implemented.
//===--------------------------------------------------------------------------
const MachineOperand * RegAllocChordal_Fer::spill(MachineOperand & mo) {
    // choose a register from the same register class as mo, and liberate
    // its use.
    const MachineOperand * sp_op = choose_register_to_spill(mo);
    this->reg_mapping->liberate_color(sp_op->getReg());

    // insert one store instruction immediatly before the definition point of
    // v_sp.
    place_store_instruction(* sp_op);

    // Dummy algorithm: insert load instructions before each instruction that
    // uses v_sp.
    place_load_instructions(* sp_op);

    // Kill the live range of v_sp, starting from the basic block where it is
    // defined
    const MachineInstr * mi = this->def_use_sites->get_def_site(* sp_op);
    const MachineBasicBlock * mbb = mi->getParent();
    this->edge_liveness->kill_live_range(* sp_op, * mbb);

    return sp_op;
}


//===--------------------------------------------------------------------------
// This method is used to find a new location to the virtual v_reg. It is
// called when the register p_reg that v_reg currently uses needs to be used to
// store a different value, while v_reg is still alive. This may  happen, for
// example, if p_reg is a precolored register, in the first instruction that
// defines p_reg. If there are other physical registers available, v_reg will
// have its value transfered to that register; otherwise, it will be spilled.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::move_or_store
                          (MachineInstr & mi, unsigned v_reg, unsigned p_reg) {
    if( this->reg_mapping->allocate_color(v_reg) ) {
        unsigned new_loc = this->reg_mapping->get_physical_location(v_reg);
        // add new_loc := p_reg before mi.
    } else {
        const MachineOperand * op =
                               this->def_use_sites->get_machine_operand(v_reg);
        place_store_instruction(*op);
        place_load_instructions(*op);
    }
}


//===--------------------------------------------------------------------------
// This method handles the definition of pre-colored registers. When a
// pre-colored register p_reg is found, the following is done: if p_reg is
// empty, simply remove it from the list of available registers. If it contains
// a virtual register v_reg, than find a new location to v_reg, and handle
// p_reg as if it was empty.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::allocate_physical_register
(unsigned p_reg, MachineInstr & mi) {
    //TODO
    std::cerr << "Allocating physical register: r" << p_reg << "\n";
    unsigned r_loc = this->reg_mapping->get_assigned_data(p_reg);
    if(r_loc == RegMapping_Fer::NO_REG) {
        // the physical register was empty. Just assign it to
        // itself.
        this->reg_mapping->allocate_color(p_reg);
    } else if(MRegisterInfo::isVirtualRegister(r_loc)) {
        // move the virtual to another place.
        this->reg_mapping->liberate_color(r_loc);
        this->reg_mapping->allocate_color(p_reg);
        move_or_store(mi, r_loc, p_reg);
    } else if(MRegisterInfo::isPhysicalRegister(r_loc)) {
        assert(r_loc == p_reg &&
              "Physical reg allocated to different physical reg.");
        // nothing else to be done: the physical register already
        // contains itself.
    }
}

//===--------------------------------------------------------------------------
// This method gets one clique of register, and one instruction. The clique is
// supposed to represent all the registers alive immediately before the
// instruction is executed. The method assigns a valid physical register to
// the virtual register. If there is not physical register available, it will
// spill one of the virtual registers, to make room for the virtual.
// Before doing the allocation, the method verify if all the virtual registers
// in the given clique are allocated. It will be an error if any of them is
// not assigned to a physical register.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::allocate_defs
                       (MachineInstr & mi, std::vector<RegClique_Fer *> & ql) {
    // First, handle the explicit definitions:
    for (unsigned i = 0; i != mi.getNumOperands(); i++) {
        MachineOperand & mo = mi.getOperand(i);
        if(mo.isDef() && mo.isRegister() && mo.getReg()) {
            unsigned reg = mo.getReg();
            // If there is no more available colors...
            if(MRegisterInfo::isVirtualRegister(reg)) {
                if( ! this->reg_mapping->allocate_color(reg) ) {
                    spill(mo);
                    this->reg_mapping->allocate_color(reg);
                }
                unsigned p_reg = reg_mapping->get_physical_location(reg);
/*
                mi.getOperand(i).setReg(p_reg);
std::cerr  << NeatPrint_Fer::print_instr(mi, * this->machine_function) << ": " << NeatPrint_Fer::print_op(reg, * this->machine_function) << " := " << NeatPrint_Fer::print_op(p_reg, * this->machine_function) << "\n";
*/
            } else {
                allocate_physical_register(reg, mi);
            }
        }
    }
    // Secondly, handle the implicit definitions:
    const TargetMachine & target_machine = this->machine_function->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const TargetInstrDescriptor &tid = tii.get(mi.getOpcode());
    for (const unsigned *implicit_def = tid.ImplicitDefs;
                                               *implicit_def; ++implicit_def) {
        unsigned p_reg = *implicit_def;
        allocate_physical_register(p_reg, mi);
    }
}


//===--------------------------------------------------------------------------
// It is possible that the code being compiled contains dead definitions, that
// is, a definition of a variable that is never used after been defined. In
// this case, the machine register used to hold the value of such definition
// can be liberated to allocate other live ranges.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::liberate_dead_defs
                                       (RegClique_Fer & q, MachineInstr & mi) {
    // Firstly, liberate the explicit dead defs:
    for (unsigned i = 0; i != mi.getNumOperands(); i++) {
        MachineOperand & mo = mi.getOperand(i);
        if (mo.isRegister() && mo.getReg() && mo.isDef()) {
            if( ! q.contains_reg(mo.getReg()) ) {
                this->reg_mapping->liberate_color(mo.getReg());
            }
        }
    }
    // Secondly, liberate the implicit dead defs:
    const TargetMachine & target_machine = this->machine_function->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const TargetInstrDescriptor &tid = tii.get(mi.getOpcode());
    for (const unsigned *implicit_def = tid.ImplicitDefs;
                                               *implicit_def; ++implicit_def) {
        unsigned p_reg = *implicit_def;
        if( ! q.contains_reg(p_reg) ) {
            this->reg_mapping->liberate_color(p_reg);
        }
    }
}


//===--------------------------------------------------------------------------
// After a virtual register is last used, the machine register assigned to it
// must be liberated, so it can be used to allocate other virtual registers.
// This method does exaclty this.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::liberate_uses(RegClique_Fer & q, MachineInstr & mi) {
    // liberate explicit uses:
    for (unsigned i = 0; i != mi.getNumOperands(); i++) {
        MachineOperand & mo = mi.getOperand(i);
        if (mo.isRegister() && mo.getReg() && mo.isUse()) {
            if( ! q.contains_reg(mo.getReg()) ) {
                this->reg_mapping->liberate_color(mo.getReg());
            }
        }
    }
    // liberate implicity uses:
    const TargetMachine & target_machine = this->machine_function->getTarget();
    const TargetInstrInfo & tii = * (target_machine.getInstrInfo());
    const TargetInstrDescriptor &tid = tii.get(mi.getOpcode());
    for (const unsigned *implicit_use = tid.ImplicitUses;
                                               *implicit_use; ++implicit_use) {
        unsigned p_reg = *implicit_use;
        if( ! q.contains_reg(p_reg) ) {
            this->reg_mapping->liberate_color(p_reg);
        }
    }

}


//===--------------------------------------------------------------------------
// This method updates the set of registers with data from the given
// instruction. Any defined register is removed from the clique, and any used
// register is added to the clique. If the used register is already in the
// clique, nothing is done.
//===--------------------------------------------------------------------------
RegClique_Fer * RegAllocChordal_Fer::update_list_of_cliques(
    MachineBasicBlock & mbb,
    RegClique_Fer & q,
    MachineInstr & mi
) {
    RegClique_Fer * ret_value = new RegClique_Fer(q);

    bool isPhi = false;
    isPhi = (mi.getOpcode() == TargetInstrInfo::PHI) ? true : false ;
    // handle explicit uses and defs
    for (unsigned i = 0; i != mi.getNumOperands(); i++) {
        MachineOperand & mo = mi.getOperand(i);
        if (mo.isUse() && mo.isRegister() && mo.getReg() && ( !isPhi ) ) {
            ret_value->add_reg(mo.getReg());
        } else if (mo.isDef() && mo.isRegister() && mo.getReg()) {
            ret_value->remove_reg(mo.getReg());
        }
    }
    // handle implicit uses and defs
    ret_value->handle_implicit_regs(mi);

    return ret_value;
}


//===--------------------------------------------------------------------------
// If liveness analysis information is available, the set of cliques in a basic 
// block can be found with a backwards linear scan. The first clique will be
// the out set of the basic block. Each time a new instruction is scanned, the
// register that it defines is removed from the clique, and any register that
// it uses, if not already in the clique, is added to it. Notice that phi
// instructions do not contribute uses for clique construction. Only registers
// defined in the phi function are taken into consideration.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::build_list_cliques
                 (std::vector<RegClique_Fer *> & cl, MachineBasicBlock & mbb) {

    const TargetMachine & target_machine = this->machine_function->getTarget();
    RegClique_Fer * q = new RegClique_Fer
                           (*this->machine_function, mbb, this->edge_liveness);
   cl.push_back( q );

    MachineBasicBlock::iterator mbbi = mbb.end();
    MachineInstr * mi; 
    unsigned i = 0;
    while (mbbi != mbb.begin()) {
        mbbi--;
        mi = mbbi;
        cl.push_back( update_list_of_cliques(mbb, *cl[i], *mi) );
        i++;
    }

}


//===--------------------------------------------------------------------------
// This method finds a mapping between virtual and machine registers in the
// basic block.
// The algorithm shown below finds a mapping between virtual registers and
// machine registers in linear time on the size of the interference graph.
// Notice that the interference graph is not really built. Also, notice that,
// due to properties of the SSA transformation, the algorithm is optimal if
// spills do not happen. In order to find an optimal assignment, the dominance
// tree is traversed using pre-order.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::perform_local_allocation(
         MachineBasicBlock & mbb
) {
    std::cerr << "**** Visiting MBB_" << mbb.getNumber() << " ****\n";
    // mark the node as already visited. Parts of the algorithm depend on
    // knowing if the node has been already visited.
    this->colored[mbb.getNumber()] = true;

    std::vector<RegClique_Fer *> ql;
    build_list_cliques(ql, mbb);

    for(int j = ql.size() - 1; j >= 0; j--) {
        std::cerr << ql[j]->to_string() << "\n";
    }

    this->reg_mapping->reset_physical_mapping(*ql[ql.size() - 1]);

std::cerr << this->reg_mapping->physical_to_string() << "\n";

    // Allocates a physical register to each virtual register
    MachineBasicBlock::iterator mbbi = mbb.begin();
    unsigned i = ql.size() - 2;
    while (mbbi != mbb.end()) {
        MachineInstr * mi;
        mi = mbbi;
std::cerr << "Parsing instruction " << NeatPrint_Fer::print_instr(*mi, * this->machine_function) << "\n";
std::cerr << this->reg_mapping->physical_to_string() << "\n";
        liberate_uses(*ql[i], *mi);
        allocate_defs(*mi, ql);
        liberate_dead_defs(*ql[i], *mi);
        i--;
        mbbi++;
    }

std::cerr << this->reg_mapping->physical_to_string() << "\n";

    // I'm a nice boy, and will free the memory ^o^
    std::vector<RegClique_Fer*>::iterator it;
    for(it = ql.begin(); it != ql.end(); it++) {
        delete *it;
    }
    ql.clear();
}


//===--------------------------------------------------------------------------
// This method converts the mapping from virtual register to physical register
// found by the perform_local_allocation method to a real assignment of
// physical registers in each operand of each instruction of the given basic
// block. Actually, it would be possible to implement this method directly in
// the perform_local_allocation function, but the separate implementation makes
// the program more readable.
//===--------------------------------------------------------------------------
void RegAllocChordal_Fer::hard_code_register_mapping(MachineBasicBlock & mbb) {
    std::cerr << "**** Visiting MBB_" << mbb.getNumber() << " ****\n";
    MachineBasicBlock::iterator mbbi = mbb.begin();
    while (mbbi != mbb.end()) {
        MachineInstr * mi;
        mi = mbbi;
        for (unsigned i = 0; i != mi->getNumOperands(); i++) {
            MachineOperand & mo = mi->getOperand(i);
            if(mo.isRegister() && mo.getReg()) {
                unsigned reg = mo.getReg();
                if(MRegisterInfo::isVirtualRegister(reg)) {
                    unsigned p_reg = reg_mapping->get_physical_location(reg);
                    mi->getOperand(i).setReg(p_reg);
                    std::string ins = NeatPrint_Fer::print_instr(*mi, * this->machine_function);
                    std::string vrg = NeatPrint_Fer::print_op(reg, * this->machine_function);
                    std::string prg = NeatPrint_Fer::print_op(p_reg, * this->machine_function);
                    std::cerr  << ins << ": " << vrg << " := " << prg << "\n";
                }
            }
        }
        mbbi++;
    }
}


//===--------------------------------------------------------------------------
// This method calls the register allocation algorithm.
//===--------------------------------------------------------------------------
bool RegAllocChordal_Fer::runOnMachineFunction(MachineFunction &mf) {

    initialize_data_structures(mf);
    std::cerr << "Register allocation via coloring of chordal graphs!\n";
    // Finds a mapping between virtual and physical registers:
    MachineBasicBlock *entry = mf.begin();
    std::set<MachineBasicBlock*> visited;
    for (df_ext_iterator<MachineBasicBlock*> dfi = df_ext_begin(entry, visited),
         end = df_ext_end(entry, visited); dfi != end; ++dfi) {
        MachineBasicBlock *mbb = *dfi;
        perform_local_allocation(*mbb);
    }

    // Applies the mapping to the operands in the instructions:
    for (MachineFunction::iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        hard_code_register_mapping(*mbb);
    }

    eliminate_phi_functions(mf);
    delete_data_structures();

    return false;
}


//===--------------------------------------------------------------------------
// This function creates a register allocator based on the coloring of
// chordal graphs.
//===--------------------------------------------------------------------------
FunctionPass *llvm::createChordalRegisterAllocator() {
  return new RegAllocChordal_Fer();
}
