//===------------- DefUseSites_Fer.cpp - Printing prototype ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it stores the definition and use sites of virtual registers.
//
// Date: May 28th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class implements a mapping of virtual registers to instructions.
// For each virtual, this class gives its definition site, which is unique,
// due to the SSA property, and all the use sites. A definition site is the
// instruction where the register is defined, and a use site is an
// instruction where the register is used. Such information allows to
// compute a weighting function used to choose the best candidate for
// spilling. The weighting function is given by the following algorithm:
//
// Instruction def_i := definition_site(v);
//
// Unsigned w := 1 + 10 ** loop_depth[def_i];
// 
// foreach ( Instruction i in use_sites(v) )
// |
// |    w := w + 1 + 10 ** loop_depth[i];
// |
// end foreach
//
// v2w(v) := w;
//
//===----------------------------------------------------------------------===//

#include <map>
#include <vector>
#include <sstream>                               // ostringstream
#include <iostream>                              // cerr

#include "llvm/CodeGen/Passes.h"                 // EdgeLivenessFernandoPassID
#include "llvm/Instructions.h"                   // MachineOperand
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineFunctionPass.h"    // MachineBasicBlock
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"

#include "llvm/fernando/DefUseSites_Fer.h"       // class definition

using namespace llvm;

static RegisterAnalysis<DefUseSites_Fer>
X("Virtual2DefUseMapping", "Virtual to def/use mapping - Fernando.");


//===--------------------------------------------------------------------------
// This method returns true if the given machine operand has already been sent
// to memory as the result of a spill operation.
//===--------------------------------------------------------------------------
bool DefUseSites_Fer::has_already_been_sent_to_memory
                                           (const MachineOperand & temporary) {
    return false;
}


//===--------------------------------------------------------------------------
// This method maps the unsigned description of a virtual register to the
// MachineOperand object that the unsigned represents. If the given number does
// not denote a virtual register operand, a run-time error will be reported.
//===--------------------------------------------------------------------------
const MachineOperand * DefUseSites_Fer::get_machine_operand(unsigned v_reg) {
    unsigned base_reg = v_reg - MRegisterInfo::FirstVirtualRegister;
    assert(this->v2op.size() > base_reg &&
                   "Invalid convertion between unsigned and machine operand.");
    return this->v2op[base_reg];
}


//===--------------------------------------------------------------------------
// Returns the instruction where the temporary is defined. This instruction is
// unique, because the target program is expected to be in the SSA-form.
//===--------------------------------------------------------------------------
const MachineInstr * DefUseSites_Fer::get_def_site(const MachineOperand & mo) {
    assert(mo.isRegister() && "Seeking def site of non-reg operand.");
    assert(mo.getReg() && "Seeking def site of non-valid register operand.");
    unsigned reg_num = mo.getReg();
    assert(MRegisterInfo::isVirtualRegister(reg_num) &&
           "Seeking def site of physical operand.");

    return this->v2def[reg_num];
}


//===--------------------------------------------------------------------------
// Returns a vector containing all the instruction that use the given virtual
// register.
//===--------------------------------------------------------------------------
const std::vector< const MachineInstr * > &
DefUseSites_Fer::get_use_sites(const MachineOperand & mo) {

    assert(mo.isRegister() && "Seeking use site of non-reg operand.");
    assert(mo.getReg() && "Seeking use site of non-valid register operand.");
    unsigned reg_num = mo.getReg();
    assert(MRegisterInfo::isVirtualRegister(reg_num) &&
           "Seeking use site of physical operand.");

    return this->v2uses[reg_num];
}


//===--------------------------------------------------------------------------
// Computes the e-th power of n, that is, n**e;
//===--------------------------------------------------------------------------
unsigned DefUseSites_Fer::pow(unsigned n, unsigned e) {
    unsigned p = 1;
    for(unsigned i = 0; i < e; i++) {
        p = p * n;
    }
    return p;
}


//===--------------------------------------------------------------------------
// Returns the spilling weight of the given temporary. Virtual registers with
// smaller weights are better candidates to be spilled. The heuristic function
// is computed according to the following algorithm:
//
// Instruction def_i := definition_site(v);
//
// Unsigned w := 1 + 10 ** loop_depth[def_i];
// 
// foreach ( Instruction i in use_sites(v) )
// |
// |    w := w + 1 + 10 ** loop_depth[i];
// |
// end foreach
//
// v2w(v) := w;
// 
//===--------------------------------------------------------------------------
const unsigned DefUseSites_Fer::get_weight(const MachineOperand & temporary) {
    const MachineInstr * def_site = get_def_site(temporary);
    const MachineBasicBlock * mbb_def = def_site->getParent();
    unsigned w = pow(10, this->loop_depth[mbb_def->getNumber()]);
    std::vector< const MachineInstr * > use_sites = get_use_sites(temporary);
    for(unsigned i = 0; i < use_sites.size(); i++) {
        const MachineInstr * use_site = use_sites[i];
        const MachineBasicBlock * mbb_use = use_site->getParent();
        w = w + pow(10, this->loop_depth[mbb_use->getNumber()]);
    }
    return w;
}


//===--------------------------------------------------------------------------
// This method maps the unsigned description of a virtual register to the
// MachineOperand object that the unsigned represents. If the given number does
// not denote a virtual register operand, a run-time error will be reported.
//===--------------------------------------------------------------------------
const unsigned DefUseSites_Fer::get_weight(unsigned v_reg) {
    const MachineOperand * temporary = get_machine_operand(v_reg);
    const MachineInstr * def_site = get_def_site(*temporary);
    const MachineBasicBlock * mbb_def = def_site->getParent();
    unsigned w = pow(10, this->loop_depth[mbb_def->getNumber()]);
    std::vector< const MachineInstr * > use_sites = get_use_sites(*temporary);
    for(unsigned i = 0; i < use_sites.size(); i++) {
        const MachineInstr * use_site = use_sites[i];
        const MachineBasicBlock * mbb_use = use_site->getParent();
        w = w + pow(10, this->loop_depth[mbb_use->getNumber()]);
    }
    return w;
}


//===--------------------------------------------------------------------------
// Free the memory, so the operating system can use it in other applications.
// TODO: it is necessary to verify if I am freeing the memory correctly.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::releaseMemory() {
    this->loop_depth.clear();
    this->v2def.clear();
    this->v2uses.clear();
    this->v2op.clear();
}


//===--------------------------------------------------------------------------
// This method visits a basic block and collects all the uses and defs present
// in this basic block.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::visit_basic_block(MachineBasicBlock & mbb) {

    // const TargetInstrInfo &tii = *target_machine->getInstrInfo();
    // const TargetInstrDescriptor &tid = tii.get(mi->getOpcode());

    MachineBasicBlock::iterator mii = mbb.begin();
    while (mii != mbb.end()) {
        MachineInstr *mi = mii++;
        for (unsigned i = 0; i != mi->getNumOperands(); ++i) {
            MachineOperand & mo = mi->getOperand(i);
            if (mo.isRegister() && mo.getReg()) {
                if(MRegisterInfo::isVirtualRegister(mo.getReg())) {
                    if (mo.isDef()) {
                        mark_def_site(mo, *mi);
                    } else {
                        add_to_use_list(mo, *mi);
                    }
                }
            }
        }
    }
}


//===--------------------------------------------------------------------------
// This method defines the definition site of an operand. It will produce a
// runtime error if the given operand is not a virtual register.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::mark_def_site(
                                     const MachineOperand & mo,
                                     const MachineInstr & mi
                                   ) {
    assert(mo.isRegister() && "Assigning def site to non-reg operand.");
    assert(mo.getReg() && "Marking non-valid register operand.");
    unsigned reg_num = mo.getReg();
    assert(MRegisterInfo::isVirtualRegister(reg_num) &&
                        "Attempt to assign a def site to a physical operand.");

    this->v2def[reg_num] = & mi;

    unsigned base_reg = reg_num - MRegisterInfo::FirstVirtualRegister;
    this->v2op[base_reg] = & mo;
}


//===--------------------------------------------------------------------------
// This method adds an instruction to the list of use sites of the given
// operand. It will produce a runtime error if the operand is not a virtual
// register.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::add_to_use_list(
                                       const MachineOperand & mo,
                                       const MachineInstr & mi
                                     ) {
    assert(mo.isRegister() && "Assigning use site to non-reg operand.");
    assert(mo.getReg() && "Marking non-valid register operand.");
    unsigned reg_num = mo.getReg();
    assert(MRegisterInfo::isVirtualRegister(reg_num) &&
           "Attempt to assign a use site to a physical operand.");

    if(this->v2uses.find(reg_num) == this->v2uses.end()) {
        std::vector< const MachineInstr * > aux;
        this->v2uses[reg_num] = aux;
    }
    std::vector< const MachineInstr * > & use_sites = this->v2uses[reg_num];
    use_sites.push_back(& mi);

}


//===--------------------------------------------------------------------------
// Prints in the standard error device all the information contained in this
// data structure.
//===--------------------------------------------------------------------------
inline void DefUseSites_Fer::dump() {

    const TargetInstrInfo & tii = *this->target_machine->getInstrInfo();
    std::map<unsigned, const MachineInstr *>::iterator d =
                                                           this->v2def.begin();
    std::map<unsigned, std::vector< const MachineInstr * > >::iterator u =
                                                          this->v2uses.begin();
    for( ; d != this->v2def.end(), u != this->v2uses.end(); d++, u++ ) {
        const MachineInstr * def_site = (*d).second;
        const MachineBasicBlock * mbb_def = def_site->getParent();
        const TargetInstrDescriptor & tid_def = tii.get(def_site->getOpcode());
        unsigned reg_num = (*d).first - MRegisterInfo::FirstVirtualRegister;

        unsigned reg_w = pow(10, this->loop_depth[mbb_def->getNumber()]);

        std::cerr << "- - - - - - - - - - - - - - - - - - - - - - - - - -\n";
        std::cerr << "Virtual Register: t" << reg_num << "\n";
        std::cerr << "Def   : " << tid_def.Name;
        std::cerr << ", Block " << mbb_def->getNumber() << "\n";

        std::vector< const MachineInstr * > u_sites = (*u).second;
        for(unsigned i = 0; i < u_sites.size(); i++) {
            const MachineInstr * u_site = u_sites[i];
            const MachineBasicBlock * mbb_u = u_site->getParent();
            const TargetInstrDescriptor & tid_u = tii.get(u_site->getOpcode());
            std::cerr << "Use   : " << tid_u.Name;
            std::cerr << ", Block " << mbb_u->getNumber() << "\n";
            reg_w = reg_w + pow(10, this->loop_depth[mbb_u->getNumber()]);
        }
        std::cerr << "Weight: " << reg_w << "\n";
    }
}


//===--------------------------------------------------------------------------
// Adds to the loop_depth vector the nested factor of the basic block mbb. The
// nested factor is computed by the LoopInfo analysis, which is available in
// the li variable.
//===--------------------------------------------------------------------------
void DefUseSites_Fer::set_loop_depth(MachineBasicBlock & mbb, LoopInfo & li) {
    const BasicBlock * bb = mbb.getBasicBlock();
    unsigned nested_factor = li.getLoopDepth(bb);
    unsigned mbb_number = mbb.getNumber();

    if(this->loop_depth.size() <= mbb_number) {
        this->loop_depth.resize(mbb_number + 1);
    }

    this->loop_depth[mbb_number] = nested_factor;
}


//===--------------------------------------------------------------------------
// runOnMachineFunction - collects the use/def sites of each register, and
// computes their weights.
//===--------------------------------------------------------------------------
bool DefUseSites_Fer::runOnMachineFunction(MachineFunction & mf) {

    LoopInfo * loop_info = & getAnalysis<LoopInfo>();
    this->target_machine = & mf.getTarget();

    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned num_v_regs = last_v_reg - MRegisterInfo::FirstVirtualRegister;
    this->v2op.assign(num_v_regs, NULL);

    for (MachineFunction::iterator mbb = mf.begin(), mbbe = mf.end();
                                                          mbb != mbbe; ++mbb) {
        visit_basic_block(*mbb);
        set_loop_depth(*mbb, *loop_info);
    }

    return false;
}
