//===----------- FernandoLiveStack.cpp - Printing prototype ---------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This file was developed by Fernando Magno Quintao Pereira, fernando at ucla
// dot edu, and it stores liveness information for the control flow graph.
//
// Date: May 28th, 2006
//
//===----------------------------------------------------------------------===//
//
// This file contains liveness information for the control flow graph.
// For each pair of basic blocks u and v, it records the set of variables
// alive between u and v. The requirements in terms of space is O(N*N*T),
// where N is the number of basic blocks, and T is the total number of
// temporaries in the control flow graph. Notice that this program only
// records information about virtual registers. Physical registers are
// not supposed to be alive between basic blocks.
//
//===----------------------------------------------------------------------===//

#ifndef EDGE_LIVENESS_FERNANDO_H
#define EDGE_LIVENESS_FERNANDO_H

#include "llvm/fernando/Matrix_Fernando.h"
#include "llvm/PassSupport.h"

namespace llvm {

    #define UNDEF_REG 65535

    class EdgeLiveness_Fernando : public MachineFunctionPass {

        public:

        /// The ID for this pass.
        const PassInfo *EdgeLivenessFernandoPassID;

        /// returns a string in the format (t1, t2, ..., tn), where ti is
        /// one of the temporaries alive in the edge that connects the
        /// machine basic block "origin" to "destiny".
        std::string to_string(
                               const MachineBasicBlock & origin,
                               const MachineBasicBlock & destiny
                    );

        /// Returns a vector containing the registers alive in the edge
        /// between the given basic blocks.
        const std::vector<bool> & get_live_variables(
                                  const MachineBasicBlock & origin,
                                  const MachineBasicBlock & destiny
                              );

        /// reporst that this pass does not change anything on the control
        /// flow graph.
        virtual void getAnalysisUsage(AnalysisUsage &AU) const {
            AU.setPreservesAll();
        }

        /// Informs the name of the pass, so it can appear in tools like
        /// llc.
        virtual const char *getPassName() const {
            return "Edge liveness analsys - Fernando.";
        }

        /// This method informs if the given temporary is alive in the edge
        /// between the blocks "origin" and "destiny".
        bool is_alive(
                       const MachineBasicBlock & origin,
                       const MachineBasicBlock & destiny,
                       const MachineOperand & temporary
                     );

        /// Informs the maximum number of temporary registers that can be
        /// alive in any time in any edge of the control flow graph. This is
        /// the number of temporaries used in the target program.
        unsigned get_num_virtual_registers() {
            return this->num_vreg;
        }

        /// This method removes the live range of v_sp. This must be done after
        /// spill is performed. The placement of load and store instructions
        /// must be performed in such a way that is always safe to remove the
        /// live range of the spilled register.
        void kill_live_range
                    (const MachineOperand & mo, const MachineBasicBlock & mbb);

        private:

        /// This variables records the maximum number of virtual registers that
        /// can be alive at any moment in any edge of the control flow graph.
        /// This is the number of temporaries (also called virtual) registers
        /// in the target program.
        unsigned num_vreg;

        /// the data structure where liveness information will be stored.
        Matrix_Fernando< std::vector<bool> > *live_sets;

        /// This object contains all the information about the register set
        /// in the target program, including the number of registers.
        const MRegisterInfo *reg_info;

        /// This vector contains the definition point of each variable.
        /// The definition point is the number of thebasic block where the
        /// variable was defined.
        std::vector<unsigned> defining_block;

        /// runOnMachineFunction - collects liveness information about each
        /// edge of the control flow graph.
        bool runOnMachineFunction(MachineFunction &mf);

        /// Removes the v_reg from the live set of all the blocks that can be
        /// reached from mbb.
        void kill_forward
                    (const MachineOperand & mo, const MachineBasicBlock & mbb);

        /// Removes the v_reg from the live set of all the blocks that have a
        /// path leading to mbb.
        void kill_backward
                    (const MachineOperand & mo, const MachineBasicBlock & mbb);

        /// adds the temporary to the live set of the edge that goes from
        /// machine basic block "origin" to "destiny".
        void set_alive(
                        const MachineBasicBlock & origin,
                        const MachineBasicBlock & destiny,
                        const MachineOperand & temporary
                      );

        /// Removes temporary from the live set of the edge connecting the
        /// block origin to the block destiny.
        void kill(
                   const MachineBasicBlock & origin,
                   const MachineBasicBlock & destiny,
                   const MachineOperand & temporary
                 );

        /// This function adds to the vector defining_block all the definition
        /// found in the given basic block.
        void point_defs(const MachineBasicBlock & block);

        /// This function scans every instruction of the basic block, and
        /// compute liveness information in a backward fashion.
        void compute_liveness(const MachineBasicBlock & block);

        /// This method allocates memory for the matrix.
        void initialize_data_structures(MachineFunction & mf);

        /// This function marks the definition site of the temporary. Because
        /// the target program is in SSA form, each temporary has a unique
        /// definition block.
        void handle_def(
                         const MachineOperand & temporary,
                         const MachineBasicBlock & block
                        );

        /// Every time a use is found, the graph is scanned backwards, and all
        /// the edges in which the used temporary is alive are marked. This
        /// function does exaclty this.
        void handle_use(
                         const MachineOperand & temporary,
                         const MachineBasicBlock & block
                        );

        /// Free the memory, so the operating system can use it in other
        /// applications. It deletes the mapping between temporaries to blocks
        /// where they are defined, and the matrix that contains liveness
        /// information.
        void releaseMemory();

    };  // end of EdgeLiveness_Fernando declaration

} // end llvm namespace

#endif
