//===----------------- Matrix_Fernando.cpp - Matrix ADT -------------------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This is a matrix (bidimensional array) abstract data type that uses
// templates to improve reusability.
//
// Date: May 27th, 2006
//
//===----------------------------------------------------------------------===//
//
// This file implements the matrix abstract data type. In order to access
// matrix elements, use the overloaded operator ().
// Usage example:
// Matrix_Fernando<int> imat(nrows, ncols);
// Matrix_Fernando<bool> bmat(nrows, ncols);
// bmat(1, 2) = true;
// std::cerr << imat(2, 3);
// Matrix< std::vector<bool> > *matrix;
// matrix = new Matrix< std::vector<bool> >(n_of_blocks, n_of_blocks);
//===----------------------------------------------------------------------===//

#ifndef MATRIX_FERNANDO_H
#define MATRIX_FERNANDO_H

#include <vector>
 
template<typename T> class Matrix_Fernando {

public:

  /// Constructor method. Creates a new matrix with nrows x ncols cells.
  Matrix_Fernando(unsigned nrows, unsigned ncols);

  /// Subscript operators often come in pairs. This is the mutator operator:
  /// It allows to change the objects stored in the matrix.
  T & operator() (unsigned i, unsigned j);

  /// This is the inspector operator. It gives read-only access to the element
  /// stored in cell (row==i, col==j)
  const T & operator() (unsigned i, unsigned j) const;

  /// Number of rows in this matrix
  unsigned nrows() const;

  /// Number of columns in this matrix
  unsigned ncols() const;

private:
  /// Internal representation of the matrix. A vector of vectors.
  std::vector<std::vector<T> > data_;
};

template<typename T> inline unsigned Matrix_Fernando<T>::nrows() const {
    return data_.size();
}

template<typename T> inline unsigned Matrix_Fernando<T>::ncols() const {
    return data_[0].size();
}

template<typename T>
inline T& Matrix_Fernando<T>::operator() (unsigned row, unsigned col) {
    assert( (row < nrows() && col < ncols()) && "Matrix bound error" );
    return data_[row][col];
}

template<typename T>
inline const T& Matrix_Fernando<T>::operator()(unsigned r, unsigned c) const {
    assert( (r < nrows() && c < ncols()) && "Matrix bound error" );
    return data_[r][c];
}

template<typename T>
Matrix_Fernando<T>::Matrix_Fernando(unsigned r, unsigned c) : data_ (r) {
    assert( (r > 0 && c > 0) && "Bad dimensions in matrix constructor");
    for (unsigned i = 0; i < r; ++i) {
        data_[i].resize(c);
    }
}

#endif
