//===-------- RegMapping_Fer.cpp - Bank of Machine Registers ---------===//
//
//              Register Allocation via Coloring of Chordal Graphs 
//
// This class represents the bank of machine registers for register allocation
// purposes.
//
// Date: May 27th, 2006
//
//===----------------------------------------------------------------------===//
//
// This class represents the bank of machine registers for register allocation
// purposes. Essentially, it marks which physical registers are being used,
// and which ones are free to use. For each physical register in use, it keeps
// track of which virtual registers they use.
//
//===----------------------------------------------------------------------===//


namespace llvm {
 
class RegMapping_Fer {

    public:

    /// Constructor method. Initializes the data structures. The machine
    /// function parameter is only used to provide initializing information,
    /// such as the number of virtual and physical registers.
    RegMapping_Fer(MachineFunction & mf);


    /// Constructor method. It produces a mapping that is a copy of
    /// the set initially given as parameter.
    RegMapping_Fer(RegMapping_Fer & rm);


    /// Destructor method. Gives back to the operating system the memory that
    /// was in use during the object's execution.
    ~RegMapping_Fer();


    /// This method assigns an empty machine register to v_reg, if one is
    /// available. If the assignment is successfully performed, the function
    /// will return the true value. False will be returned if there is no
    /// register available. If v_reg is already allocated to a machine
    /// register, the method will return true, and nothing else is done.
    bool allocate_color(unsigned v_reg);


    /// This method clears the machine register that has been assigned to
    /// reg. It has the effect of spilling reg, and after it is called,
    /// any register from the same class as reg will be able to be allocated.
    /// That is, after liberate_color is called, allocate_color
    /// will return true if the parameters are from the same class of
    /// registers.
    void liberate_color(unsigned reg);


    /// This method returns the virtual register that was assigned to the given
    /// physical register. If the physical register does not contain any
    /// virtual register, the method will return egMapping_Fer::NO_REG. It is
    /// possible that the given physical register is a pre-colored register.
    /// In this case, the physical value itself will be returned.
    unsigned get_assigned_data(unsigned p_reg);


    /// This method returns the physical register that was allocated to the
    /// virtual register. If the virtual register has not been allocated to
    /// any register, it will return egMapping_Fer::NO_REG.
    unsigned get_physical_location(unsigned v_reg);


    /// This method cleans the mapping between physical registers and
    /// virtual registers, and if the registers in the given clique have
    /// already been allocated, new values will be assigned to the bank of
    /// machine registers to reflect the allocation. This method is generally
    /// called before starting the allocation of a basic block.
    void reset_physical_mapping(RegClique_Fer & ql);


    /// This method returns a string in the format "t1(r1), t2(r2), ... tm(rm),
    /// where virtual registers ti are allocated in machine registers ri.
    std::string virtual_to_string();


    /// This method returns a string in the format "r1(t1), r2(t2), ... rm(tm),
    /// where physical registers ri is holding the value of virtual register
    /// ti.
    std::string physical_to_string();


    /// This constant is used to identify which physical registers are free,
    /// and can be assigned to virtual registers.
    static const unsigned NO_REG = 65535;

    private:

    /// This method determines if p_reg is not currently allocated to any
    /// virtual register. In order to verify this, it does not suffice to
    /// check if the cell of p_reg in the physical register table contains the
    /// NO_REG value, because physical registers contains alias. For instance,
    /// in the X86 architecture, the physical registers AL, AX and EAX are all
    /// alias, and if one of them is in use, the others are not free. If p_reg
    /// is not been used to allocate any virtual, the method returns true. It
    /// returns false otherwise.
    bool is_register_available(unsigned p_reg);


    /// This method returns a register than can be assigned to the given
    /// virtual register. It will return RegMapping_Fer::NO_REG if there
    /// is no register available.
    unsigned get_free_register(unsigned v_reg);


    /// Free the memory, so the operating system can use it in other
    /// applications.
    void releaseMemory();


    /// This is the table that maps physical registers to virtual registers.
    /// If p2v[p_reg] = NO_REG, than p_reg is not assigned to any
    /// virtual register.
    std::vector<unsigned> p2v;


    /// This vector maps virtual registers to physical registers. A
    /// BankOfRegisters_Fer::NO_REG value indicates that the virtual register
    /// hasn't been assigned a value yet.
    std::vector<unsigned> v2p;


    /// The bank of register keeps a reference to the machine function in
    /// order to get the architecture specific information.
    MachineFunction * machine_function;

};


//===--------------------------------------------------------------------------
// Constructor method. Initializes the data structures. The machine function
// parameter is only used to provide initializing information, such as the
// number of virtual and physical registers.
//===--------------------------------------------------------------------------
RegMapping_Fer::RegMapping_Fer(MachineFunction & mf) {
    const TargetMachine & target_machine = mf.getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    unsigned num_p_regs = reg_info->getNumRegs();
// TODO verify why RegMapping_Fer::NO_REG cannot be used here...
    this->p2v.assign(num_p_regs, 65535);
    SSARegMap * srm = mf.getSSARegMap();
    unsigned last_v_reg = srm->getLastVirtReg() + 1;
    unsigned num_v_regs = last_v_reg - MRegisterInfo::FirstVirtualRegister;
    this->v2p.assign(num_v_regs, 65535);
    this->machine_function = & mf;
}


//===--------------------------------------------------------------------------
// Constructor method. It produces a mapping that is a copy of the set
// initially given as parameter.
//===--------------------------------------------------------------------------
RegMapping_Fer::RegMapping_Fer(RegMapping_Fer & rm) {
    this->p2v.assign(rm.p2v.begin(), rm.p2v.end());
    this->v2p.assign(rm.v2p.begin(), rm.v2p.end());
    this->machine_function = rm.machine_function;
}


//===--------------------------------------------------------------------------
// Destructor method. Gives back to the operating system the memory that was in
// use during the object's execution.
//===--------------------------------------------------------------------------
RegMapping_Fer::~RegMapping_Fer() {
    releaseMemory();
}


//===--------------------------------------------------------------------------
// This method determines if p_reg is not currently allocated to any virtual
// register. In order to verify this, it does not suffice to check if the
// cell of p_reg in the physical register table contains the NO_REG value,
// because physical registers contains alias. For instance, in the X86
// architecture, the physical registers AL, AX and EAX are all alias, and if
// one of them is in use, the others are not free. If p_reg is not been
// used to allocate any virtual, the method returns true. It returns false
// otherwise.
//===--------------------------------------------------------------------------
bool RegMapping_Fer::is_register_available(unsigned p_reg) {

    if (this->p2v[p_reg] != RegMapping_Fer::NO_REG) {
        return false;
    }

    const TargetMachine & target_machine = this->machine_function->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();

    for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
        if (this->p2v[*a] != RegMapping_Fer::NO_REG) {
            return false;
        }
    }

    return true;
}


//===--------------------------------------------------------------------------
// This method returns the virtual register that was assigned to the given
// physical register. If the physical register does not contain any virtual
// register, the method will return egMapping_Fer::NO_REG.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_assigned_data(unsigned p_reg) {
    assert(p_reg < this->p2v.size() && "Invalid physical register.");
    unsigned data = this->p2v[p_reg];
    if(data == RegMapping_Fer::NO_REG) {
        const TargetMachine & target_machine = machine_function->getTarget();
        const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
        for (const unsigned *a = reg_info->getAliasSet(p_reg); *a; ++a) {
            if (this->p2v[*a] != RegMapping_Fer::NO_REG) {
                data = this->p2v[*a];
                break;
            }
        }
    }
    return data;
}


//===--------------------------------------------------------------------------
// This method returns the physical register that was assigned to the virtual
// register. If the virtual register has not been allocated to any register, it
// will return null. It is also possible that the given parameter describes a
// physical register. In this case, it is a pre-colored register, and its
// location must be the physical register itself.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_physical_location(unsigned reg) {
    if(MRegisterInfo::isVirtualRegister(reg)) {
        unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
        assert(base_reg < this->v2p.size() && "Invalid virtual register.");
        return this->v2p[base_reg];
    } else {
        if(this->p2v[reg] != RegMapping_Fer::NO_REG) {
            assert(this->p2v[reg] == reg &&
                               "Pre-colored register not assigned to itself.");
        }
        return this->p2v[reg];
    }
}


//===--------------------------------------------------------------------------
// This method returns a register that can be assigned to the given virtual
// register. It will return RegMapping_Fer.NO_REG if there is no register
// available.
//===--------------------------------------------------------------------------
unsigned RegMapping_Fer::get_free_register(unsigned v_reg) {
    const TargetRegisterClass *trc =
                    this->machine_function->getSSARegMap()->getRegClass(v_reg);
    TargetRegisterClass::iterator trci =
                         trc->allocation_order_begin(* this->machine_function);
    TargetRegisterClass::iterator trce =
                           trc->allocation_order_end(* this->machine_function);

    for (; trci != trce; ++trci) {
        if (is_register_available(*trci)) {
            return *trci;
        }
    }

    return RegMapping_Fer::NO_REG;
}

//===--------------------------------------------------------------------------
// This method clears the machine register that is assigned to the virtual
// v_reg. It has the effect of spilling v_reg, and after it is called, any
// register from the same class as v_reg will be able to be allocated. That is,
// after liberate_color is called, allocate_color will return true if the
// parameters are from the same class of registers.
//===--------------------------------------------------------------------------
void RegMapping_Fer::liberate_color(unsigned reg) {
    if(MRegisterInfo::isVirtualRegister(reg)) {
        unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
        unsigned p_reg = this->v2p[base_reg];
        if(p_reg != RegMapping_Fer::NO_REG) {
            if(this->p2v[p_reg] == reg) {
                this->p2v[p_reg] = RegMapping_Fer::NO_REG;
            }
        }
    } else {
        this->p2v[reg] = RegMapping_Fer::NO_REG;
    }
}


//===--------------------------------------------------------------------------
// This method assigns an empty machine register to reg, if one is available.
// If the assignment is successfully performed, the function will return the
// true value. False will be returned if there is no register available.
// If reg is already allocated to a machine register, the method will return
// true, and nothing else is done.
//===--------------------------------------------------------------------------
bool RegMapping_Fer::allocate_color(unsigned reg) {

    bool is_v = MRegisterInfo::isVirtualRegister(reg);
    bool is_p = MRegisterInfo::isPhysicalRegister(reg);
    assert( (is_v || is_p) && "Unknow register specification.");

    if(is_v) {
        unsigned p_reg = 0;
        unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
        if(this->v2p[base_reg] != RegMapping_Fer::NO_REG) {
            assert(this->p2v[this->v2p[base_reg]] == reg &&
                                             "Inconsistent register mapping.");
            // the virtual already has a location :)
        } else {
            p_reg = get_free_register(reg);
            if(p_reg == RegMapping_Fer::NO_REG) {
                // there was no free register :(
                return false;
            } else {
                this->p2v[p_reg] = reg;
                this->v2p[base_reg] = p_reg;
            }
        }
    } else if(is_p) {
        assert(this->p2v[reg] == RegMapping_Fer::NO_REG &&
                                    "Overwriting value in physical register.");
        this->p2v[reg] = reg;
    }
    return true;
}


//===--------------------------------------------------------------------------
// Free the memory, so the operating system can use it in other applications.
// TODO: it is necessary to verify if I am freeing the memory correctly.
//===--------------------------------------------------------------------------
void RegMapping_Fer::releaseMemory() {
    this->p2v.clear();
    this->v2p.clear();
}


//===--------------------------------------------------------------------------
// This method cleans the mapping between physical registers and virtual
// registers, and if the registers in the given clique have already been
// allocated, new values will be assigned to the bank of machine registers to
// reflect the allocation. This method is generally called before starting the
// allocation of a basic block.
//===--------------------------------------------------------------------------
void RegMapping_Fer::reset_physical_mapping(RegClique_Fer & ql) {
    for(unsigned u = 0; u < this->p2v.size(); u++) {
        this->p2v[u] = RegMapping_Fer::NO_REG;
    }
    std::vector<unsigned> new_alloc;
    ql.to_vector(new_alloc);
    for(unsigned u = 0; u < new_alloc.size(); u++) {
        unsigned reg = new_alloc[u];
        if(MRegisterInfo::isVirtualRegister(reg)) {
            unsigned base_reg = reg - MRegisterInfo::FirstVirtualRegister;
            if(this->v2p[base_reg] != RegMapping_Fer::NO_REG) {
                this->p2v[this->v2p[base_reg]] = reg;
            }
        } else {
            this->p2v[reg] = reg;
        }
    }
}


//===--------------------------------------------------------------------------
// This method returns a string in the format "t1(r1), t2(r2), ... tm(rm),
// where virtual registers ti are allocated in machine registers ri.
//===--------------------------------------------------------------------------
std::string RegMapping_Fer::virtual_to_string() {
    std::ostringstream aux;
    bool is_first = true;
    const TargetMachine & target_machine = this->machine_function->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for(unsigned i = 0; i < this->v2p.size(); i++) {
        if(this->v2p[i] != RegMapping_Fer::NO_REG) {
            const char * phys_reg_name = reg_info->getName(this->v2p[i]);
            if(is_first) {
                aux << "t" << i << "(" << phys_reg_name << ")";
                is_first = false;
            } else {
                aux << ", " << "t" << i << "(" << phys_reg_name << ")";
            }
        }
    }
    return aux.str();
}


//===--------------------------------------------------------------------------
// This method returns a string in the format "r1(t1), r2(t2), ... rm(tm),
// where physical registers ri is holding the value of virtual register ti.
//===--------------------------------------------------------------------------
std::string RegMapping_Fer::physical_to_string() {
    std::ostringstream aux;
    bool is_first = true;
    const TargetMachine & target_machine = this->machine_function->getTarget();
    const MRegisterInfo * reg_info = target_machine.getRegisterInfo();
    for(unsigned i = 0; i < this->p2v.size(); i++) {
        if(this->p2v[i] != RegMapping_Fer::NO_REG) {
            unsigned reg = this->p2v[i];
            std::ostringstream reg_name;
            if(MRegisterInfo::isVirtualRegister(reg)) {
                reg_name << "t" << (reg - MRegisterInfo::FirstVirtualRegister);
            } else {
                reg_name << reg_info->getName(i);
            }
            const char * phys_reg_name = reg_info->getName(i);
            std::ostringstream num_name;
            num_name << "r" << i << "/" << phys_reg_name;
            if(is_first) {
                aux << num_name.str() << "(" << reg_name.str() << ")";
                is_first = false;
            } else {
                aux << ", " << num_name.str() << "(" << reg_name.str() << ")";
            }
        }
    }
    return aux.str();
}


} // end of llvm namespace
